/*Si scriva un programma che esegua le seguenti operazioni:

- crea e utilizza una zona di memoria condivisa
- attiva un timer (funzione di sistema "setitimer") che ogni 20s invia un segnale (SIGALARM)
- alla ricezione del segnale disattiva il timer e attiva un processo "child" che legge un file (a piacere), lo stiva nella memoria condivisa, invia un segnale al processo "parent" e poi termina.
- il processo "parent" stampa il contenuto della memoria condivisa sul terminale e riattiva il timer
*/
#include "header.h"
#include <sys/time.h>

key_t key;

struct data {
int flag;
int data;
char testo[1024*1024];
} *shm_ptr;

struct itimerval temp, *timer;
struct timeval transf;

void sighandlertime(int code);
void sighandlerparent(int code);
void assegnatempo(int sec);

int main(int argc, char *argv[])
{
int t, fp, pid;
char testo;

if (argc!=2) {
   printf("Hai dimenticato di specificare il file da leggere.\n");
   exit(-1);
}
signal(SIGALRM,sighandlertime);
signal(SIGUSR1,sighandlerparent);
key=ftok("/home/vladislav/tc3/",1);
shm_ptr = ShmCreate(key, sizeof(struct data), 0666, 0);   // get a shared memory segment
if (!shm_ptr) {
   perror("Cannot create shared memory");
   exit(-1);
}
assegnatempo(5);
setitimer(ITIMER_REAL,timer,0);
while (1) {
      pause();
      if ((pid = fork()) < 0) { // on error exit
         printf("Error on child creation, %s\n", strerror(errno));
         exit(-1);
      }
      if (pid==0) { //figlio
         if (shm_ptr->flag==0) {
            shm_ptr->flag=1;                              // lock shared memory
            if ((fp=open(argv[1],O_RDONLY))==-1) { //aperto solo in lettura
               perror("open input");
               exit(-1);
            }
            for (t=0; read(fp,&testo,1)!=0 && t<1048576; t++) shm_ptr->testo[t]=testo;
            shm_ptr->data=t;              //salvo numero dati letto
            shm_ptr->flag=0;                          // unlock shared memory
         }
         kill(getppid(), SIGUSR1);    //invia segnale al padre
         return 0;
      }
      else { //padre
         if (shm_ptr->flag==0) {
            shm_ptr->flag=1;                              // lock shared memory
            printf("parent\n");
            for (t=0;t<shm_ptr->data;t++) printf("%c",shm_ptr->testo[t]);
            shm_ptr->flag=0;                          // unlock shared memory
         }
         assegnatempo(20);
         setitimer(ITIMER_REAL,timer,0);
      }
}
return 0;
}

void assegnatempo(int sec)
{
transf.tv_sec=sec;
transf.tv_usec=0;
temp.it_interval=transf;
temp.it_value=transf;
timer=&temp;
return;
}

void sighandlertime(int code)
{
  printf("\ngot signal %d\n",code);
  assegnatempo(0);
  setitimer(ITIMER_REAL,timer,0);
  return;
}

void sighandlerparent(int code)
{
  printf("\ngot signal %d\n",code);
  return;
}

/*
 * Function ShmCreate:
 * Create and attach a SysV shared memory segment to the current process.
 *
 * First call get a shared memory segment with KEY key access and size SIZE,
 * by creating it with R/W privilege for the user (this is the meaning of
 * the ored flags). The function return an identifier shmid used for any 
 * further reference to the shared memory segment. 
 * Second call attach the shared memory segment to this process and return a
 * pointer to it (of void * type). 
 * Then initialize shared memory to the given value
 *
 * Input:  an IPC key value
 *         the shared memory segment size
 *         the permissions
 *         the fill value
 * Return: the address of the shared memory segment (NULL on error)
 */

void * ShmCreate(key_t ipc_key, int shm_size, int perm, int fill) 
{
    void * shm_ptr;
    int shm_id;                       // ID of the IPC shared memory segment 
    shm_id = shmget(ipc_key, shm_size, IPC_CREAT|perm);        // get shm ID
    if (shm_id < 0) { 
	return NULL;
    }
    shm_ptr = shmat(shm_id, NULL, 0);                  // map it into memory
    if (shm_ptr < 0) {
	return NULL;
    }
    memset((void *)shm_ptr, fill, shm_size);                 // fill segment 
    return shm_ptr;
}

/*
 * Function ShmFind:
 * Find a SysV shared memory segment 
 * Input:  an IPC key value
 *         the shared memory segment size
 * Return: the address of the segment (NULL on error)
 */

void * ShmFind(key_t ipc_key, int shm_size) 
{
    void * shm_ptr;
    int shm_id;                      // ID of the SysV shared memory segment 
    shm_id = shmget(ipc_key, shm_size, 0);          // find shared memory ID 
    if (shm_id < 0) {
	return NULL;
    }
    shm_ptr = shmat(shm_id, NULL, 0);                  // map it into memory 
    if (shm_ptr < 0) {
	return NULL;
    }
    return shm_ptr;
}

/*
 * Function ShmRemove:
 * Schedule removal for a SysV shared memory segment 
 * Input:  an IPC key value
 *         the shared memory segment size
 * Return: 0 on success, -1 on error
 */

int ShmRemove(key_t ipc_key, void * shm_ptr) 
{
    int shm_id;                      // ID of the SysV shared memory segment 
    //first detach segment
    if (shmdt(shm_ptr) < 0) {
	return -1;
    }
    //schedule segment removal 
    shm_id = shmget(ipc_key, 0, 0);                 // find shared memory ID
    if (shm_id < 0) {
	if (errno == EIDRM) return 0;
	return -1;
    }
    if (shmctl(shm_id, IPC_RMID, NULL) < 0) {             // ask for removal
	if (errno == EIDRM) return 0;
	return -1;
    }
    return 0;
}
